@@ -1,10 +1,8 @@
 /* window.h -- Structure and flags used in manipulating Info windows.
-   $Id: window.h,v 1.11 2008/06/28 08:10:02 gray Exp $
+   $Id: window.h 5884 2014-10-22 22:19:04Z gavin $
 
-   This file is part of GNU Info, a program for reading online documentation
-   stored in Info format.
-
-   Copyright (C) 1993, 1997, 2004, 2007 Free Software Foundation, Inc.
+   Copyright 1993, 1997, 2004, 2007, 2011 2013, 2014
+   Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -19,13 +17,14 @@
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
-   Written by Brian Fox (bfox@ai.mit.edu). */
+   Originally written by Brian Fox. */
 
 #ifndef INFO_WINDOW_H
 #define INFO_WINDOW_H
 
-#include "infomap.h"
+#include "doc.h"
 #include "nodes.h"
+#include <regex.h>
 
 /* Smallest number of visible lines in a window.  The actual height is
    always one more than this number because each window has a modeline. */
@@ -40,6 +39,7 @@
    point values into columns on screen and vice versa. */
 typedef struct line_map_struct
 {
+  NODE *node;      /* Node to which this line pertains */
   size_t nline;    /* Line number for which the map is computed. */
   size_t size;     /* Number of elements map can accomodate */
   size_t used;     /* Number of used map slots */
@@ -56,9 +56,13 @@ typedef struct line_map_struct
    do it.  NB> The last element does NOT end with a semi-colon. */
 #define WINDOW_STATE_DECL \
    NODE *node;          /* The node displayed in this window. */ \
-   int pagetop;         /* LINE_STARTS[PAGETOP] is first line in WINDOW. */ \
+   long pagetop;        /* LINE_STARTS[PAGETOP] is first line in WINDOW. */ \
    long point           /* Offset within NODE of the cursor position. */
 
+typedef struct {
+  WINDOW_STATE_DECL;            /* What gets saved. */
+} WINDOW_STATE;
+
 /* Structure which defines a window.  Windows are doubly linked, next
    and prev. The list of windows is kept on WINDOWS.  The structure member
    window->height is the total height of the window.  The position location
@@ -69,31 +73,34 @@ typedef struct window_struct
 {
   struct window_struct *next;      /* Next window in this chain. */
   struct window_struct *prev;      /* Previous window in this chain. */
-  int width;            /* Width of this window. */
-  int height;           /* Height of this window. */
-  int first_row;        /* Offset of the first line in the_screen. */
-  int goal_column;      /* The column we would like the cursor to appear in. */
-  Keymap keymap;        /* Keymap used to read commands in this window. */
+  long width;           /* Width of this window. */
+  long height;          /* Height of this window. */
+  long first_row;       /* Offset of the first line in the_screen. */
+  long goal_column;     /* Column to place the cursor in when moving it up and 
+                           down.  -1 means the column it is currently in. */
   WINDOW_STATE_DECL;    /* Node, pagetop and point. */
   LINE_MAP line_map;    /* Current line map */
   char *modeline;       /* Calculated text of the modeline for this window. */
-  char **line_starts;   /* Array of printed line starts for this node. */
-  int line_count;       /* Number of lines appearing in LINE_STARTS. */
+  long *line_starts;    /* Offsets of printed line starts in node->contents.*/
+  long *log_line_no;    /* Number of logical line corresponding to each
+                           physical one. */
+  long line_count;      /* Number of elements in LINE_STARTS and LOG_LINE_NO.*/
+  size_t line_slots;    /* Allocated space in LINE_STARTS and LOG_LINE_NO. */
+
   int flags;            /* See below for details. */
+
+  /* Used for highlighting search matches. */
+  char *search_string;
+  int search_is_case_sensitive;
+  regmatch_t *matches;
+  size_t match_count;
+
+  /* History of nodes visited in this window. */
+  WINDOW_STATE **hist;       /* Nodes visited in this window. */
+  size_t hist_index;            /* Index where to add the next node. */
+  size_t hist_slots;            /* Number of slots allocated to HIST. */
 } WINDOW;
 
-typedef struct {
-  WINDOW_STATE_DECL;            /* What gets saved. */
-} WINDOW_STATE;
-
-/* Structure defining the current state of an incremental search. */
-typedef struct {
-  WINDOW_STATE_DECL;    /* The node, pagetop and point. */
-  int search_index;     /* Offset of the last char in the search string. */
-  int direction;        /* The direction that this search is heading in. */
-  int failing;          /* Whether or not this search failed. */
-} SEARCH_STATE;
-
 #define W_UpdateWindow  0x01    /* WINDOW needs updating. */
 #define W_WindowIsPerm  0x02    /* This WINDOW is a permanent object. */
 #define W_WindowVisible 0x04    /* This WINDOW is currently visible. */
@@ -123,24 +130,15 @@ extern void window_make_modeline (WINDOW *window);
    You pass WIDTH and HEIGHT; the dimensions of the total screen size. */
 extern void window_initialize_windows (int width, int height);
 
-/* Make a new window showing NODE, and return that window structure.
-   The new window is made to be the active window.  If NODE is passed
-   as NULL, then show the node showing in the active window.  If the
-   window could not be made return a NULL pointer.  The active window
-   is not changed.*/
-extern WINDOW *window_make_window (NODE *node);
+/* Make a new window by splitting an existing one. If the window could
+   not be made return a null pointer.  The active window is not changed .*/
+extern WINDOW *window_make_window (void);
 
 /* Delete WINDOW from the list of known windows.  If this window was the
    active window, make the next window in the chain be the active window,
    or the previous window in the chain if there is no next window. */
 extern void window_delete_window (WINDOW *window);
 
-/* A function to call when the screen changes size, and some windows have
-   to get deleted.  The function is called with the window to be deleted
-   as an argument, and it can't do anything about the window getting deleted;
-   it can only clean up dangling references to that window. */
-extern VFunction *window_deletion_notifier;
-
 /* Set WINDOW to display NODE. */
 extern void window_set_node_of_window (WINDOW *window, NODE *node);
 
@@ -155,6 +153,8 @@ extern void window_new_screen_size (int width, int height);
    window, then no change takes place. */
 extern void window_change_window_height (WINDOW *window, int amount);
 
+extern void set_window_pagetop (WINDOW *window, int desired_top);
+
 /* Adjust the pagetop of WINDOW such that the cursor point will be visible. */
 extern void window_adjust_pagetop (WINDOW *window);
 
@@ -178,92 +178,60 @@ extern void window_unmark_chain (WINDOW *chain, int flag);
 /* Make WINDOW start displaying at PERCENT percentage of its node. */
 extern void window_goto_percentage (WINDOW *window, int percent);
 
-/* Build a new node which has FORMAT printed with ARG1 and ARG2 as the
+/* Build a new node which has AP printed according to FORMAT as the
    contents. */
-extern NODE *build_message_node (const char *format, void *arg1, void *arg2);
+extern NODE *build_message_node (const char *format, va_list ap);
 
-/* Useful functions can be called from outside of window.c. */
-extern void initialize_message_buffer (void);
+extern NODE *format_message_node (const char *format, ...)
+  TEXINFO_PRINTFLIKE(1,2);
 
-/* Print FORMAT with ARG1,2 to the end of the current message buffer. */
-extern void printf_to_message_buffer (const char *format, void *arg1, void *arg2,
-    void *arg3);
+struct text_buffer;
+extern NODE *text_buffer_to_node (struct text_buffer *tb);
 
-/* Convert the contents of the message buffer to a node. */
-extern NODE *message_buffer_to_node (void);
+/* Make a message appear in the echo area, built from arguments formatted
+   according to FORMAT.
 
-/* Return the length of the most recently printed line in message buffer. */
-extern int message_buffer_length_this_line (void);
-
-/* Pad STRING to COUNT characters by inserting blanks. */
-extern int pad_to (int count, char *string);
-
-/* Make a message appear in the echo area, built from FORMAT, ARG1 and ARG2.
-   The arguments are treated similar to printf () arguments, but not all of
-   printf () hair is present.  The message appears immediately.  If there was
+   The message appears immediately.  If there was
    already a message appearing in the echo area, it is removed. */
-extern void window_message_in_echo_area (const char *format, void *arg1, void *arg2);
+extern void window_message_in_echo_area (const char *format, ...)
+  TEXINFO_PRINTFLIKE(1,2);
 
-/* Place a temporary message in the echo area built from FORMAT, ARG1
-   and ARG2.  The message appears immediately, but does not destroy
+extern void vwindow_message_in_echo_area (const char *format, va_list ap);
+
+extern void free_echo_area (void);
+
+/* Place a temporary message in the echo area built from arguments
+   formatted as per FORMAT.
+
+   The message appears immediately, but does not destroy
    any existing message.  A future call to unmessage_in_echo_area ()
    restores the old contents. */
-extern void message_in_echo_area (const char *format, void *arg1, void *arg2);
+extern void message_in_echo_area (const char *format, ...)
+  TEXINFO_PRINTFLIKE(1,2);
+
 extern void unmessage_in_echo_area (void);
 
 /* Clear the echo area, removing any message that is already present.
    The echo area is cleared immediately. */
 extern void window_clear_echo_area (void);
 
-/* Quickly guess the approximate number of lines to that NODE would
-   take to display.  This really only counts carriage returns. */
-extern int window_physical_lines (NODE *node);
-
-/* Calculate a list of line starts for the node belonging to WINDOW.  The line
-   starts are pointers to the actual text within WINDOW->NODE. */
-extern void calculate_line_starts (WINDOW *window);
-
-/* Given WINDOW, recalculate the line starts for the node it displays. */
-extern void recalculate_line_starts (WINDOW *window);
-
-/* Return the number of characters it takes to display CHARACTER on the
-   screen at HPOS. */
-extern int character_width (int character, int hpos);
-
-/* Return the number of characters it takes to display STRING on the
-   screen at HPOS. */
-extern int string_width (char *string, int hpos);
-
 /* Return the index of the line containing point. */
 extern int window_line_of_point (WINDOW *window);
 
-/* Get and return the goal column for this window. */
-extern int window_get_goal_column (WINDOW *window);
-
 /* Get and return the printed column offset of the cursor in this window. */
 extern int window_get_cursor_column (WINDOW *window);
 
-/* Get and Set the node, pagetop, and point of WINDOW. */
-extern void window_get_state (WINDOW *window, SEARCH_STATE *state);
-extern void window_set_state (WINDOW *window, SEARCH_STATE *state);
-
-/* Count the number of characters in current line of WIN that precede
-   the printed column offset of GOAL. */
-extern int window_chars_to_goal (WINDOW *win, int goal);
-
 extern size_t process_node_text
-        (WINDOW *win, char *start, int do_tags,
-         int (*fun) (void *, size_t, const char *, char *, size_t, size_t),
-	 void *closure);
-
-void clean_manpage (char *manpage);
+        (WINDOW *win, char *start,
+         int (*fun) (WINDOW *, size_t, size_t, size_t, char *,
+		     size_t, size_t));
 
 extern void window_compute_line_map (WINDOW *win);
 
-int window_point_to_column (WINDOW *win, long point, long *np);
+extern int window_point_to_column (WINDOW *win, long point, long *np);
 
-void window_line_map_init (WINDOW *win);
+extern void window_line_map_init (WINDOW *win);
 
-long window_end_of_line (WINDOW *win);
+extern long window_log_to_phys_line (WINDOW *window, long ln);
 
 #endif /* not INFO_WINDOW_H */
